# Generated by Django 3.0.8 on 2020-08-03 04:12

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0018_auto_20200802_1619'),
    ]

    operations = [
        migrations.AlterField(
            model_name='emailotpauthentication',
            name='key',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='otp_authentications', to='avishan.Email'),
        ),
        migrations.AlterField(
            model_name='phoneotpauthentication',
            name='key',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='otp_authentications', to='avishan.Phone'),
        ),
    ]



# Generated by Django 3.0.7 on 2020-06-25 13:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0013_auto_20200622_1536'),
    ]

    operations = [
        migrations.RenameField(
            model_name='emailpasswordauthenticate',
            old_name='email',
            new_name='key',
        ),
        migrations.RenameField(
            model_name='emailverification',
            old_name='email',
            new_name='identifier',
        ),
        migrations.RenameField(
            model_name='phoneotpauthenticate',
            old_name='phone',
            new_name='key',
        ),
        migrations.RenameField(
            model_name='phonepasswordauthenticate',
            old_name='phone',
            new_name='key',
        ),
        migrations.RenameField(
            model_name='phoneverification',
            old_name='phone',
            new_name='identifier',
        ),
        migrations.AlterField(
            model_name='emailverification',
            name='verification_code',
            field=models.CharField(default=None, max_length=255),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='phoneverification',
            name='verification_code',
            field=models.CharField(default=None, max_length=255),
            preserve_default=False,
        ),
    ]



import inspect
import json
from typing import List, get_type_hints, Type, Callable, Optional, Union

from django.core.handlers.wsgi import WSGIRequest
from django.db.models import QuerySet
from django.http import JsonResponse
from django.http.response import HttpResponseBase
from django.utils import timezone
from django.views import View

from avishan.configure import get_avishan_config
from avishan.descriptor import DirectCallable, FunctionAttribute, DjangoAvishanModel
from avishan.exceptions import ErrorMessageException, AvishanException, AuthException
from avishan.misc import status
from avishan.misc.translation import AvishanTranslatable
from avishan.models import AvishanModel, RequestTrack
from avishan.middlewares import AvishanRequestStorage


# todo fix cors motherfucker
# todo doc
class AvishanView(View):
    authenticate: bool = True
    track_it: bool = False
    is_api: bool = None

    def setup(self, request, *args, **kwargs):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        super().setup(request, *args, **kwargs)
        self.response: dict = request.avishan.response
        self.request: WSGIRequest
        self.class_attributes_type = get_type_hints(self.__class__)

        for kwarg_key, kwarg_value in kwargs.items():
            if hasattr(self, kwarg_key):
                self.__setattr__(kwarg_key, kwarg_value)

        for kwarg_key in request.GET.keys():
            kwarg_key: str
            if hasattr(self, kwarg_key):
                data = request.GET.getlist(kwarg_key)
                if len(data) > 1:
                    parsed = []
                    for item in data:
                        parsed.append(self.cast_data(kwarg_key, item))
                    self.__setattr__(kwarg_key, parsed)
                else:
                    self.__setattr__(kwarg_key, self.cast_data(kwarg_key, data[0]))

        request.avishan.view_class = self
        request.avishan.on_error_view_class = self

        request.avishan.is_api = self.is_api
        if self.track_it and not request.avishan.is_tracked:
            request.avishan.is_tracked = True
            request.avishan.request_track_object = RequestTrack.objects.create()

    def http_method_not_allowed(self, request, *args, **kwargs):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        request.avishan.status_code = status.HTTP_405_METHOD_NOT_ALLOWED
        self.response['allowed_methods'] = self.get_allowed_methods()
        return super().http_method_not_allowed(request, *args, **kwargs)

    def get_allowed_methods(self) -> List[str]:
        allowed = []
        for method in ['get', 'post', 'put', 'delete']:
            if callable(getattr(self, method, False)):
                allowed.append(method)
        return allowed

    def dispatch(self, request, *args, **kwargs):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        if request.avishan.exception:
            return

        try:
            if self.authenticate and not self.is_authenticated(request):
                raise AuthException(AuthException.TOKEN_NOT_FOUND)
            request.avishan.view_start_time = timezone.now()
            result = super().dispatch(request, *args, **kwargs)
            request.avishan.view_end_time = timezone.now()

        except AvishanException as e:
            raise e
        except Exception as e:
            AvishanException(wrap_exception=e)
            raise e
        if request.avishan.exception:
            return
        return result

    def is_authenticated(self, request) -> bool:
        """
        Checks for user available in storage
        :return: true if authenticated
        """
        if not request.avishan.authentication_object:
            return False
        return True

    def cast_data(self, key, value):
        if key in self.class_attributes_type.keys():
            if isinstance(self.class_attributes_type[key], type):
                return self.class_attributes_type[key](value)
            return self.class_attributes_type[key].__args__[0](value)
        return value


class AvishanApiView(AvishanView):
    is_api = True
    authenticate = True

    def dispatch(self, request, *args, **kwargs):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        if request.method not in ['GET', 'DELETE']:
            try:
                if len(request.body) > 0:
                    request.data = json.loads(request.body.decode('utf-8'))
                else:
                    request.data = {}
            except:
                request.data = {}

        super().dispatch(request, *args, **kwargs)
        if request.avishan.can_touch_response:
            return JsonResponse(self.response)
        return self.response


class AvishanTemplateView(AvishanView):
    is_api = False
    template_file_address: str = None
    template_url: str = None
    context: dict = {}

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.context = {
            **self.context,
            **{
                'CURRENT_REQUEST': request,
                'AVISHAN_CONFIG': get_avishan_config(),
                'self': self
            }
        }

    def dispatch(self, request, *args, **kwargs):
        self.parse_request_post_to_data(request)

        return super().dispatch(request, *args, **kwargs)

    def parse_request_post_to_data(self, request):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        if request.method in ['POST', 'PUT']:
            try:
                if len(request.body) > 0:
                    request.data = dict(request.POST)
                    for key, value in request.data.items():
                        if len(value) == 1:
                            request.data[key] = value[0]
                else:
                    request.data = {}
            except:
                request.data = {}

    def get(self, request, *args, **kwargs):
        return self.render()

    def render(self):
        from django.shortcuts import render as django_render
        return django_render(self.request, self.template_file_address, self.context)


class AvishanModelApiView(AvishanApiView):
    track_it = True
    model: Type[AvishanModel] = None
    model_item: AvishanModel = None
    model_function: Callable = None  # todo these sends model not dict
    model_function_name: str = None
    model_descripted: DjangoAvishanModel = None
    direct_callable: Optional[DirectCallable] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setup(self, request, *args, **kwargs):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        super().setup(request, *args, **kwargs)
        model_plural_name = kwargs.get('model_plural_name', None)
        model_item_id = kwargs.get('model_item_id', None)
        self.model_function_name = kwargs.get('model_function_name', None)
        self.model_descripted: DjangoAvishanModel = request.avishan.project.find_model(snake_case_plural_name=model_plural_name)
        if not self.model_descripted:
            raise ErrorMessageException('Entered model name not found')
        self.model = self.model_descripted.target

        if self.model_function_name is None:
            if request.method == 'GET':
                if model_item_id is None:
                    self.model_function_name = 'all'
                else:
                    self.model_function_name = 'get'
            elif request.method == 'POST':
                self.model_function_name = 'create'
            elif request.method == 'PUT':
                self.model_function_name = 'update'
            elif request.method == 'DELETE':
                self.model_function_name = 'remove'

        if model_item_id is not None:
            self.model_item = self.model.get(avishan_raise_400=True, id=int(model_item_id))
        if self.model_function_name is not None:
            for direct_callable in self.model_descripted.methods:
                if self.model_function_name == direct_callable.name:
                    self.direct_callable = direct_callable
                    break
            if not self.direct_callable:
                raise AuthException(AuthException.METHOD_NOT_DIRECT_CALLABLE)
            self.authenticate = self.direct_callable.authenticate
            try:
                if self.model_item is None:
                    self.model_function = getattr(self.model, self.model_function_name)
                else:
                    self.model_function = getattr(self.model_item, self.model_function_name)
            except AttributeError:
                raise ErrorMessageException(AvishanTranslatable(
                    EN=f'Requested method not found in model {self.model.class_name()}'
                ))

    def get(self, request, *args, **kwargs):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        if self.model_function_name == 'get':
            result = self.model.get(id=self.model_item.id)
        else:
            result = self.model_function()

            if isinstance(result, QuerySet):
                result = self.model.queryset_handler(request.GET, queryset=result)
        response = self.parse_returned_data(request, result)
        if request.avishan.can_touch_response:
            self.response[self.direct_callable.response_json_key] = response
        else:
            self.response = response

    def post(self, request, *args, **kwargs):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        request_data = self.parse_request_data(**request.data)
        result = self.model_function(**request_data)
        response = self.parse_returned_data(request, result)
        if request.avishan.can_touch_response:
            self.response[self.direct_callable.response_json_key] = response
        else:
            self.response = response

    def put(self, request, *args, **kwargs):
        self.post(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        result = self.model_function()
        response = self.parse_returned_data(request, result)
        if request.avishan.can_touch_response:
            self.response[self.direct_callable.response_json_key] = response
        else:
            self.response = response

    @staticmethod
    def parse_returned_data(request, returned):
        # noinspection PyTypeHints
        request.avishan: AvishanRequestStorage

        if isinstance(returned, HttpResponseBase):
            request.avishan.can_touch_response = False
            return returned
        if isinstance(returned, QuerySet):
            return [item.to_dict() for item in returned]
        elif isinstance(returned, list):
            if len(returned) > 0 and isinstance(returned[0], AvishanModel):
                return [item.to_dict() for item in returned]
            return returned
        elif isinstance(returned, AvishanModel):
            return returned.to_dict()
        else:
            return returned

    def parse_request_data(self, **kwargs):
        """Parse Request Body

        Request body consists of multiple items, some have primitive data, like name:str or price:float. But for
        relational objects like birth_place:City, we should convert json-acceptable data to specific python objects. In
        this case we should get corresponding object from database.
        """

        cleaned = {}
        for function_attribute in self.direct_callable.args:
            function_attribute: FunctionAttribute

            """Errors for required data before python raises exception for function params"""
            if function_attribute.is_required and function_attribute.name not in kwargs.keys():
                raise ErrorMessageException(f'field "{function_attribute.name}" is required')

            """Optional field not provided"""
            if function_attribute.name not in kwargs.keys():
                continue

            """Objects and array of objects should convert to db objects"""
            if function_attribute.type is function_attribute.TYPE.OBJECT:
                cleaned[function_attribute.name] = self._parse_request_object(
                    function_attribute=function_attribute, target_dict=kwargs[function_attribute.name]
                )
            elif function_attribute.type is function_attribute.TYPE.ARRAY and \
                    inspect.isclass(function_attribute.type_of) and \
                    issubclass(function_attribute.type_of, AvishanModel):
                cleaned[function_attribute.name] = [
                    self._parse_request_object(
                        function_attribute=function_attribute, target_dict=item
                    ) for item in kwargs[function_attribute.name]
                ]
            elif function_attribute.type is function_attribute.TYPE.FILE:
                raise NotImplementedError()
            else:
                """Casts primitive data here"""
                cleaned[function_attribute.name] = function_attribute.type_caster(
                    entry=kwargs[function_attribute.name],
                    target_type=function_attribute.type
                )

        return cleaned

    @staticmethod
    def _parse_request_object(function_attribute: FunctionAttribute, target_dict: dict):
        """Parse object

        Get objects from database

        :param FunctionAttribute function_attribute: corresponding attribute data
        :param dict target_dict: request data
        :return: cleaned object
        """

        if not isinstance(target_dict, dict):
            if target_dict is None:
                return None
            raise ErrorMessageException(f'value for "{function_attribute.name}" must be dict consist of id or other '
                                        f'unique values so that db can find corresponding object')

        return function_attribute.type_of.request_arg_get_from_dict(target_dict)

class Redoc(AvishanTemplateView):
    template_file_address = 'avishan/redoc.html'
    authenticate = False
    track_it = False

    def get(self, request, *args, **kwargs):
        from avishan.libraries.openapi3 import OpenApi

        open_api_yaml = OpenApi().export_yaml()

        text_file = open('static/openapi.yaml', 'w+')
        text_file.write(open_api_yaml)
        text_file.close()
        return self.render()



from typing import Optional, Union, List

from .misc import status
from .misc.translation import AvishanTranslatable
from crum import get_current_request


class AvishanException(Exception):
    def __init__(
            self,
            wrap_exception: Optional[Exception] = None,
            status_code: int = status.HTTP_400_BAD_REQUEST
    ):
        save_traceback()
        request = get_current_request()
        if wrap_exception:
            if isinstance(wrap_exception, KeyError):
                body = f'field {wrap_exception.args[0]} not found in data and its required'
            else:
                if len(wrap_exception.args) == 0:
                    body = wrap_exception.__class__.__name__
                elif len(wrap_exception.args) == 1:
                    body = str(wrap_exception.args[0])
                else:
                    body = str(wrap_exception.args)[1:-1]
            request.avishan.exception = wrap_exception
            request.avishan.status_code = status.HTTP_418_IM_TEAPOT
            add_error_message_to_response(
                body=body,
            )
        else:
            request.avishan.exception = self
            request.avishan.status_code = status_code


class AuthException(AvishanException):
    from .misc.translation import AvishanTranslatable
    """
    Error kinds
    """
    NOT_DEFINED = 0, AvishanTranslatable(EN='Not Defined', FA='مشخص نشده')
    ACCOUNT_NOT_FOUND = 1, AvishanTranslatable(EN='User Account not found', FA='حساب کاربری پیدا نشد')
    ACCOUNT_NOT_ACTIVE = 2, AvishanTranslatable(EN='Deactivated User Account', FA='حساب کاربری غیرفعال است')
    GROUP_ACCOUNT_NOT_ACTIVE = 3, AvishanTranslatable(
        EN='User Account Deactivated in Selected User Group',
        FA='حساب کاربری در گروه‌کاربری انتخاب شده غیر فعال است'
    )
    TOKEN_NOT_FOUND = 4, AvishanTranslatable(EN='Token not found', FA='توکن پیدا نشد')
    TOKEN_EXPIRED = 5, AvishanTranslatable(EN='Token timed out', FA='زمان استفاده از توکن تمام شده است')
    ERROR_IN_TOKEN = 6, AvishanTranslatable(EN='Error in token', FA='خطا در توکن')
    ACCESS_DENIED = 7, AvishanTranslatable(EN='Access Denied', FA='دسترسی غیرمجاز')
    HTTP_METHOD_NOT_ALLOWED = 8, AvishanTranslatable(EN='HTTP method not allowed in this url')
    INCORRECT_PASSWORD = 9, AvishanTranslatable(EN='Incorrect Password', FA='رمز اشتباه است')
    DUPLICATE_AUTHENTICATION_IDENTIFIER = 10, AvishanTranslatable(
        EN='Authentication Identifier already Exists',
        FA='شناسه احراز هویت تکراری است'
    )
    DUPLICATE_AUTHENTICATION_TYPE = 11, AvishanTranslatable(
        EN='Duplicate Authentication Type for User Account',
        FA='روش احراز هویت برای این حساب کاربری تکراری است'
    )
    DEACTIVATED_TOKEN = 12, AvishanTranslatable(
        EN='Token Deactivated, Sign in again',
        FA='توکن غیرفعال شده است، دوباره وارد شوید'
    )
    MULTIPLE_CONNECTED_ACCOUNTS = 13, AvishanTranslatable(
        EN='Multiple Accounts found with this identifier.',
        FA='چند حساب با این شناسه پیدا شد'
    )

    METHOD_NOT_DIRECT_CALLABLE = 14, AvishanTranslatable(
        EN='Method is not callable direct to model',
        FA='تابع به طور مستقیم قابل صدا زدن نیست'
    )

    PASSWORD_NOT_FOUND = 15, AvishanTranslatable(
        EN='Password not found for user account',
        FA='رمز برای این کاربر تنظیم نشده است'
    )

    def __init__(self, error_kind: tuple = NOT_DEFINED):
        from .misc.translation import AvishanTranslatable
        status_code = status.HTTP_403_FORBIDDEN
        self.error_kind = error_kind
        if error_kind[0] == AuthException.HTTP_METHOD_NOT_ALLOWED[0]:
            status_code = status.HTTP_405_METHOD_NOT_ALLOWED
        super().__init__(status_code=status_code)
        add_error_message_to_response(code=error_kind[0], body=str(error_kind[1]), title=str(AvishanTranslatable(
            EN='Authentication Exception',
            FA='خطای احراز هویت'
        )))

    @classmethod
    def get_login_required_errors(cls) -> List[tuple]:
        return [
            cls.ACCOUNT_NOT_FOUND,
            cls.ACCOUNT_NOT_ACTIVE,
            cls.GROUP_ACCOUNT_NOT_ACTIVE,
            cls.TOKEN_NOT_FOUND,
            cls.TOKEN_EXPIRED,
            cls.ERROR_IN_TOKEN,
            cls.INCORRECT_PASSWORD,
            cls.DEACTIVATED_TOKEN,
            cls.MULTIPLE_CONNECTED_ACCOUNTS
        ]


class ErrorMessageException(AvishanException):
    def __init__(self, message: Union[str, AvishanTranslatable] = 'Error',
                 status_code: int = status.HTTP_400_BAD_REQUEST):
        super().__init__(status_code=status_code)
        message = str(message)
        # add_error_message_to_response(
        #     body=message if message else AvishanTranslatable(
        #         EN='Error details not provided',
        #         FA='توضیحات خطا ارائه نشده').__str__(),
        #     title=AvishanTranslatable(
        #         EN='Error',
        #         FA='خطا'
        #     ).__str__()
        # )
        add_error_message_to_response(
            body=message if message else 'Error details not provided',
            title='Error'
        )


def add_debug_message_to_response(body: str = None, title: str = None):
    debug = {}
    if body is not None:
        debug['body'] = body
    if title is not None:
        debug['title'] = title
    get_current_request().avishan.messages['debug'].append(debug)


def add_info_message_to_response(body: str = None, title: str = None):
    info = {}
    if body is not None:
        info['body'] = body
    if title is not None:
        info['title'] = title
    get_current_request().avishan.messages['info'].append(info)


def add_success_message_to_response(body: str = None, title: str = None):
    success = {}
    if body is not None:
        success['body'] = body
    if title is not None:
        success['title'] = title
    get_current_request().avishan.messages['success'].append(success)


def add_warning_message_to_response(body: str = None, title: str = None):
    warning = {}
    if body is not None:
        warning['body'] = body
    if title is not None:
        warning['title'] = title
    get_current_request().avishan.messages['warning'].append(warning)


def add_error_message_to_response(body: str = None, title: str = None, code=None):
    error = {}
    if body is not None:
        error['body'] = body
    if title is not None:
        error['title'] = title
    if code is not None:
        error['code'] = code
    get_current_request().avishan.messages['error'].append(error)


def save_traceback():
    request = get_current_request()
    try:
        if request.avishan.traceback is not None:
            return
    except KeyError:
        return
    import sys, traceback
    exc_type, exc_value, exc_tb = sys.exc_info()
    tbe = traceback.TracebackException(
        exc_type, exc_value, exc_tb,
    )
    if tbe.exc_traceback is not None:
        request.avishan.traceback = ''.join(tbe.format())
        if request.avishan.debug:
            print(request.avishan.traceback)



import inspect
import datetime
from enum import Enum, auto
from typing import List, Callable, Optional, Tuple, Union

import docstring_parser
import stringcase
from django.apps import apps
from django.conf import settings
from django.db import models
from django.db.models.base import ModelBase
from djmoney.models.fields import MoneyField
from docstring_parser import DocstringMeta, DocstringParam

from avishan.misc import status


class Project:
    def __init__(self, name: str):
        self.name: str = name
        self.apps: List[DjangoApplication] = self.load_apps()

        a = 1

    def load_apps(self) -> List['DjangoApplication']:
        from avishan.configure import get_avishan_config
        return [DjangoApplication(project=self, name=app_name) for app_name in settings.INSTALLED_APPS if
                app_name not in get_avishan_config().descriptor_ignored_installed_apps()]

    def models_pool(self) -> List['DjangoAvishanModel']:
        total = []
        for app in self.apps:
            total.extend(app.models)
        return total

    def find_model(self, name: str = None, plural_name: str = None, snake_case_plural_name: str = None) -> Optional[
        'DjangoAvishanModel']:
        for model in self.models_pool():
            if (name and model.name == name) or \
                    (plural_name and model.target.class_plural_name() == plural_name) or \
                    (snake_case_plural_name and model.target.class_plural_name() == stringcase.pascalcase(
                        snake_case_plural_name)):
                return model

    def __str__(self):
        return self.name


class DjangoApplication:
    def __init__(self, project: Project, name: str):
        self.project: Project = project
        self.name: str = name
        self.models: List[DjangoAvishanModel] = self.load_models()

    def load_models(self) -> List['DjangoAvishanModel']:
        total = []
        for model in apps.get_app_config(self.name).get_models():
            total.append(DjangoAvishanModel(self, model))
        return total

    def __str__(self):
        return self.name


# todo django model needed for raw models

class DataModel:
    def __init__(self,
                 name: str,
                 attributes: List['Attribute'],
                 description: str = None
                 ):
        self.name = name
        self.attributes = attributes
        self.description = description


class DjangoAvishanModel(DataModel):
    """
    Model descriptor for django avishan models. Not tested for django models.
    """

    def __init__(self, app: DjangoApplication, target):
        from avishan.models import AvishanModel

        self.target: AvishanModel = target
        super().__init__(
            name=target._meta.object_name,
            attributes=self.extract_attributes(),
            description=self.target._model_description()
        )
        self.app: DjangoApplication = app
        self.methods = self.extract_methods()
        self.parse_auto_resolves()
        self.prepare_docs()

    @property
    def direct_callables(self) -> List['DirectCallable']:
        return self.methods

    def prepare_docs(self):
        document_added_methods = []
        for method_name, method in inspect.getmembers(self.target, predicate=inspect.isfunction) + \
                                   inspect.getmembers(self.target, predicate=inspect.ismethod):
            if not method_name.endswith('_documentation_raw'):
                continue
            document_added_methods.append(method_name[1:-18])
        for method_name in document_added_methods:
            doc = getattr(self.target, f'_{method_name}_documentation_raw')()
            if hasattr(self.target, f'_{method_name}_documentation_params'):
                doc %= getattr(self.target, f'_{method_name}_documentation_params')()
            if inspect.ismethod(getattr(self.target, method_name)):
                getattr(self.target, method_name).__func__.__doc__ = doc
            else:
                getattr(self.target, method_name).__doc__ = doc

    def extract_attributes(self) -> List['DjangoFieldAttribute']:
        """
        Extracts fields from model and create DjangoFieldAttribute from them.
        """
        # todo should not be from openapi
        return [DjangoFieldAttribute(target=self.target.get_field(field_name)) for field_name in
                self.target.openapi_documented_fields()]

    def extract_methods(self) -> List['DirectCallable']:
        """
        Extracts methods from model and create Method from them.
        """
        # todo extract methods here, not only apis
        return self.target.direct_callable_methods()

    def parse_auto_resolves(self):
        """auto resolve request body"""
        for item in self.methods:
            if item.documentation is None:
                continue
            if item.documentation.request_body is not None \
                    and isinstance(item.documentation.request_body.attributes,
                                   RequestBodyDocumentation.AutoResolveRequestBody):
                item.documentation.request_body.attributes = item.args

            if item.authenticate and item.documentation.response_bodies:
                item.documentation.response_bodies.append(
                    ResponseBodyDocumentation(
                        title='Authentication error',
                        status_code=status.HTTP_403_FORBIDDEN
                    )
                )
            if item.documentation.request_body:
                related_names = []
                for attribute in item.documentation.request_body.attributes:
                    if attribute.type is Attribute.TYPE.OBJECT:
                        related_names.append(attribute.type_of.class_name())
                if len(related_names) > 0:
                    item.documentation.response_bodies.append(
                        ResponseBodyDocumentation(
                            title=f'Related object of one of these types not found: {", ".join(related_names)}',
                            status_code=status.HTTP_404_NOT_FOUND
                        )
                    )

    def is_abstract(self) -> bool:
        return self.target._meta.abstract

    def __str__(self):
        return self.name


class Function:

    def __init__(self, target):
        self.target = target
        self.name = target.__name__
        self.short_description: Optional[str] = None
        self.long_description: Optional[str] = None
        self.args: List[Attribute] = self.load_args_from_signature(self.target)
        self.returns: Optional[Attribute] = None
        self._doc = docstring_parser.parse(target.__doc__)
        if self.__class__ == Function:
            self.load_from_doc()

    def load_from_doc(self):
        """
        Caution: Override args only when doc have data about them. else keep signature data
        :return:
        :rtype:
        """
        raise NotImplementedError()

    @classmethod
    def load_args_from_signature(cls, method) -> List['Attribute']:
        args = []
        for key, value in dict(inspect.signature(method).parameters.items()).items():
            if key in ['self', 'cls', 'kwargs', 'avishan_raise_400']:
                continue
            args.append(FunctionAttribute(value))
        return args

    def __str__(self):
        return self.name


class Method(Function):

    def __init__(self, target):
        super().__init__(target)
        self.model = self.get_class_that_defined_method(target)
        self.is_class_method = inspect.ismethod(target)
        if self.__class__ == Method:
            self.load_from_doc()

    @staticmethod
    def get_class_that_defined_method(method):
        if inspect.ismethod(method):
            for cls in inspect.getmro(method.__self__.__class__):
                if cls.__dict__.get(method.__name__) is method:
                    return cls
            method = method.__func__
        if inspect.isfunction(method):
            cls = getattr(inspect.getmodule(method),
                          method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
            if isinstance(cls, type):
                return cls
        return getattr(method, '__objclass__', None)


class ApiMethod(Method):
    class METHOD(Enum):
        GET = auto()
        POST = auto()
        PUT = auto()
        DELETE = auto()

    class RESPONSE:
        def __init__(self, returns: 'Attribute', description: str, status_code: int = 200, ):
            self.status_code = status_code
            self.returns = returns
            self.description = description

    def __init__(self, target, url: str, method: 'ApiMethod.METHOD'):
        super().__init__(target)
        self.responses: List[ApiMethod.RESPONSE] = []
        self.method: ApiMethod.METHOD = method
        self.url: str = url
        # self.load_from_doc()

    def load_from_doc(self):
        self.short_description = self._doc.short_description
        self.long_description = self._doc.long_description
        if len(self._doc.params) > 0:
            self.args = [Attribute.create_from_doc_param(item) for item in self._doc.params]
        self.responses = self.load_responses_from_doc()

    def load_responses_from_doc(self) -> List['ApiMethod.RESPONSE']:
        responses = []
        for item in self._doc.meta:
            if not isinstance(item, DocstringMeta):
                continue
            if item.args[0] != 'response':
                continue
            responses.append(
                ApiMethod.RESPONSE(
                    returns=Attribute(
                        name='returns',
                        type=Attribute.create_type_from_doc_param(item.args[1]),
                        type_of=Attribute.create_type_of_from_doc_param(item.args[1])
                    ),
                    description=item.description,
                    status_code=int(item.args[2])
                )
            )
        return responses


class DirectCallable(ApiMethod):

    def __init__(self,
                 model: type,
                 target_name: str,
                 response_json_key: str,
                 url: str = None,
                 method: ApiMethod.METHOD = ApiMethod.METHOD.GET,
                 authenticate: bool = True,
                 hide_in_redoc: bool = False,
                 is_class_method: bool = None,
                 documentation: 'ApiDocumentation' = None
                 ):
        from avishan.configure import get_avishan_config
        from avishan.models import AvishanModel
        model: AvishanModel

        if url is None:
            auto_set_url = True
            url = f'/{target_name}'
        else:
            auto_set_url = False

        super().__init__(
            target=getattr(model, target_name),
            url='/' + get_avishan_config().AVISHAN_URLS_START + f'/{model.class_plural_snake_case_name()}' + url,
            method=method
        )

        self.model = model
        self.authenticate = authenticate
        self.hide_in_redoc = hide_in_redoc
        self.response_json_key = response_json_key
        if is_class_method is not None:
            self.is_class_method = is_class_method

        if not self.is_class_method and auto_set_url:
            self.url = '/' + get_avishan_config().AVISHAN_URLS_START + f'/{model.class_plural_snake_case_name()}/' \
                       + '{id}' + url

        self.documentation = documentation

    def __str__(self):
        return f'{self.model.class_name()}.{self.name}'


class Attribute:
    """
    Defining Attributes details
    """

    class TYPE(Enum):
        """
        Types reduced to this standard and globally acceptable values
        """
        STRING = auto()
        INT = auto()
        FLOAT = auto()
        DATE = auto()
        TIME = auto()
        DATETIME = auto()
        BOOLEAN = auto()
        OBJECT = auto()
        ARRAY = auto()
        FILE = auto()
        DATA_MODEL = auto()

    """
    Type pool for checking targets
    """
    _TYPE_POOL = {
        TYPE.STRING: (str, models.CharField, models.TextField),
        TYPE.INT: (int, models.IntegerField, models.AutoField),
        TYPE.FLOAT: (float, models.FloatField, MoneyField),
        TYPE.DATE: (datetime.date, models.DateField),
        TYPE.TIME: (datetime.time, models.TimeField),
        TYPE.DATETIME: (datetime.datetime, models.DateTimeField),
        TYPE.BOOLEAN: (bool, models.BooleanField),
        TYPE.OBJECT: (models.OneToOneField, models.ForeignKey),
        TYPE.ARRAY: (list, List, tuple, Tuple),
        TYPE.FILE: (models.FileField,),
        TYPE.DATA_MODEL: ()
    }

    NO_DEFAULT = 'NO_DESCRIPTOR_DEFAULT'

    def __init__(self,
                 name: str,
                 type: 'Attribute.TYPE',
                 type_of: Union[type, DataModel, 'AvishanModel'] = None,
                 default=NO_DEFAULT,
                 description: str = None,
                 example: str = None,
                 is_required: bool = True,
                 choices: List[str] = None
                 ):
        self.name = name
        self.type: Attribute.TYPE = type
        """
        If representation type is OBJECT or ARRAY it can be defined.
        """
        self.type_of = type_of
        self.default = default
        self.description = description
        self.example = example
        self.is_required = is_required
        self.choices = choices
        self._doc = None

    @classmethod
    def create_from_doc_param(cls, doc_param: DocstringParam) -> 'Attribute':
        return Attribute(
            name=doc_param.arg_name,
            type=Attribute.create_type_from_doc_param(doc_param.type_name),
            type_of=Attribute.create_type_of_from_doc_param(doc_param.type_name),
            description=doc_param.description,
            is_required=not doc_param.is_optional,
            default=doc_param.default
        )

    @classmethod
    def create_type_from_doc_param(cls, type_string: str) -> 'Attribute.TYPE':
        if type_string.find('[') > 0:
            return Attribute.type_finder(entry=type_string[:type_string.find('[')])

        return Attribute.type_finder(entry=type_string)

    @classmethod
    def create_type_of_from_doc_param(cls, type_string: str) -> Optional[type]:
        from avishan.models import AvishanModel

        if type_string.find('[') > 0:
            found = cls.create_type_from_doc_param(type_string=type_string)
            if found is Attribute.TYPE.ARRAY:
                return AvishanModel.get_model_with_class_name(type_string[type_string.find('[') + 1: -1])
        if isinstance(type_string, str):
            return AvishanModel.get_model_with_class_name(type_string)
        return None

    @staticmethod
    def type_finder(entry) -> 'Attribute.TYPE':
        from avishan.models import AvishanModel
        if entry == 'datetime.datetime':
            entry = datetime.datetime
        elif entry == 'datetime.date':
            entry = datetime.date
        elif entry == 'datetime.time':
            entry = datetime.time

        try:
            from typing import _Union
            if isinstance(entry, _Union):
                for item in entry.__args__:
                    return Attribute.type_finder(item)
        except ImportError:
            pass

        for target, pool in Attribute._TYPE_POOL.items():
            for swimmer in pool:
                if entry is swimmer:
                    return target

        # instanced type
        for target, pool in Attribute._TYPE_POOL.items():
            for swimmer in pool:
                if swimmer is str:
                    continue
                if isinstance(entry, swimmer):
                    return target

        # inherited type
        for target, pool in Attribute._TYPE_POOL.items():
            for swimmer in pool:
                if inspect.isclass(entry) and issubclass(entry, swimmer):
                    return target

        # string of type
        for target, pool in Attribute._TYPE_POOL.items():
            for swimmer in pool:
                if entry == swimmer.__name__:
                    return target

        if (isinstance(entry, str) and AvishanModel.get_model_with_class_name(entry) is not None) \
                or isinstance(entry, ModelBase):
            return Attribute.TYPE.OBJECT

        raise NotImplementedError()

    @staticmethod
    def type_caster(entry, target_type: 'Attribute.TYPE'):
        from avishan.models import AvishanModel

        if entry is None:
            return None

        if target_type is Attribute.TYPE.STRING:
            cast_class = str
        elif target_type is Attribute.TYPE.INT:
            cast_class = int
        elif target_type is Attribute.TYPE.FLOAT:
            cast_class = float
        elif target_type is Attribute.TYPE.DATETIME:
            cast_class = datetime.datetime
        elif target_type is Attribute.TYPE.DATE:
            cast_class = datetime.date
        elif target_type is Attribute.TYPE.BOOLEAN:
            cast_class = bool
        else:
            return entry

        return AvishanModel.cast_type_data(cast_class, entry)

    def __str__(self):
        return self.name


class DjangoFieldAttribute(Attribute):

    # todo default
    def __init__(self, target: models.Field):
        default = self.NO_DEFAULT
        if target.has_default():
            if isinstance(target.get_default(), Enum):
                default = target.get_default().value
            else:
                default = target.get_default()

        choices = None
        if target.choices:
            choices = []
            for value, label in target.choices:
                choices.append(label)

        super().__init__(
            name=target.name,
            type=self.define_representation_type(target),
            description=target.help_text,
            default=default,
            is_required=self.check_is_required(target),
            choices=choices
        )
        if self.type is Attribute.TYPE.OBJECT:
            if isinstance(target, (models.OneToOneField, models.ForeignKey)):
                self.type_of = target.related_model
            else:
                raise NotImplementedError()
        if self.type is Attribute.TYPE.ARRAY:
            raise NotImplementedError()
        if self.type is Attribute.TYPE.FILE:
            self.type = Attribute.TYPE.STRING
        self.target = target

    @classmethod
    def check_is_required(cls, target) -> bool:
        """
        Checks if field is required
        """
        if hasattr(target.model, 'is_field_required'):
            assert isinstance(target.model.is_field_required, Callable)
            return target.model.is_field_required(target)
        raise NotImplementedError()

    @staticmethod
    def define_representation_type(field: models.Field) -> Attribute.TYPE:
        """
        converts django fields models to defined types.
        :param field: target field
        """
        return Attribute.type_finder(field.__class__)


class FunctionAttribute(Attribute):

    # todo default
    def __init__(self, parameter: inspect.Parameter):
        from avishan.models import AvishanModel

        if parameter is inspect.Parameter.empty or parameter.name in ['self', 'cls']:
            raise ValueError

        default = self.NO_DEFAULT
        if parameter.default is not inspect.Parameter.empty:
            default = parameter.default
        super().__init__(
            name=parameter.name,
            type=self.define_representation_type(parameter),
            default=default
        )
        temp = parameter.annotation
        if self.type is Attribute.TYPE.OBJECT:
            if isinstance(parameter.annotation, str):
                self.type_of = AvishanModel.get_model_with_class_name(parameter.annotation)
            else:
                try:
                    from typing import _Union
                    if isinstance(parameter.annotation, _Union):
                        for item in parameter.annotation.__args__:
                            temp = item
                            break
                except Exception:
                    pass
                self.type_of = temp
        if self.type is Attribute.TYPE.ARRAY:
            temp = parameter.annotation.__args__[0]
            found_here = False
            try:
                from typing import _ForwardRef
                if isinstance(temp, _ForwardRef):
                    temp = AvishanModel.get_model_with_class_name(temp.__forward_arg__)
                    found_here = True
            except Exception:
                pass

            if not found_here:
                temp = Attribute.type_finder(temp)
            self.type_of = temp
        self.is_required = self.define_is_required(parameter)

    @staticmethod
    def define_representation_type(parameter: inspect.Parameter) -> Attribute.TYPE:
        return Attribute.type_finder(parameter.annotation)

    @staticmethod
    def define_is_required(parameter: inspect.Parameter):
        if parameter.default is not inspect.Parameter.empty:
            return False
        return True


class RequestBodyDocumentation:
    class AutoResolveRequestBody:
        pass

    class Example:
        def __init__(self,
                     name: str,
                     summary: str,
                     value: dict
                     ):
            self.name = name
            self.summary = summary
            self.value = value

    def __init__(self,
                 attributes: Union[List[Attribute], AutoResolveRequestBody] = None,
                 description: str = None,
                 examples: List[Example] = None):
        if attributes is None:
            attributes = []
        if examples is None:
            examples = []
        self.attributes = attributes
        self.description = description
        self.examples = examples


class ResponseBodyDocumentation:
    class AutoResolveResponseBody:
        pass

    def __init__(self,
                 title: str,
                 status_code: int = status.HTTP_200_OK,
                 attributes: Union[List[Attribute], AutoResolveResponseBody] = None,
                 description: str = None
                 ):
        if attributes is None:
            attributes = []
        self.status_code = status_code
        self.attributes = attributes
        self.title = title
        self.description = description


class ApiDocumentation:
    def __init__(self,
                 title: str,
                 description: Optional[str] = None,
                 request_body: RequestBodyDocumentation = None,
                 response_bodies: List[ResponseBodyDocumentation] = None,
                 direct_callable: DirectCallable = None,
                 ):
        if response_bodies is None:
            response_bodies = []
        self.direct_callable = direct_callable
        self.title = title
        self.description = description
        self.request_body = request_body
        self.response_bodies = response_bodies

    def __str__(self):
        return str(self.direct_callable)

# todo path, not single api can have global params: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#parameterObject
# todo add tag for requests & tag description
# todo can describe model itself. check pet_model in redoc example
# todo example for request data



from typing import Union, Type, List

from avishan.descriptor import Project
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import BaseCommand


# WARNING: no import from avishan here

class AvishanConfigure:

    @staticmethod
    def command_line_checkup(management_command):
        # todo
        # AvishanConfigure.check_for_config_file(management_command)
        AvishanConfigure.check_django_settings(management_command)

    @staticmethod
    def check_avishan_config_file():
        project_config = None
        for sub_class in AvishanConfigFather.__subclasses__():
            if project_config is not None:
                raise ValueError('Multiple implementation of AvishanConfigFather class found')
            project_config = sub_class

        if project_config is None:
            raise ValueError('AvishanConfig class not found. Run "python manage.py avishan_configure" command')

    @staticmethod
    def create_avishan_config_file():
        f = open('avishan_config.py', 'w+')

        f.writelines((
            'class AvishanConfig(AvishanConfigFather):\n',
            "    pass\n"
        ))
        f.close()

    @staticmethod
    def check_for_config_file(management_command: BaseCommand = None):
        from avishan.configure import AvishanConfigFather, AvishanConfigure

        project_config = None
        for sub_class in AvishanConfigFather.__subclasses__():
            if project_config is not None:
                message = 'Multiple implementation of AvishanConfigFather class found'
                management_command.stdout.write(management_command.style.ERROR(message))
                return
            project_config = sub_class

        if not project_config:
            AvishanConfigure.create_avishan_config_file()
            message = 'Successfully created config file. run again to check it.'
            management_command.stdout.write(management_command.style.SUCCESS(message))
        else:
            AvishanConfigure.check_avishan_config_file()

    @staticmethod
    def check_django_settings(management_command):
        from avishan.utils import find_file
        print(find_file('settings.py', '.'), '*')


class AvishanConfigFather:
    class LANGUAGES:
        FA = 'FA'
        EN = 'EN'

    PROJECT: Project = None
    PROJECT_NAME: str = None
    MONITORED_APPS_NAMES: List[str] = []
    NOT_MONITORED_STARTS: List[str] = ['/admin', '/static', '/media', '/favicon.ico', '/api/av1/redoc']
    IGNORE_TRACKING_STARTS: List[str] = []
    AVISHAN_URLS_START = 'api/av1'
    JWT_KEY: str = None
    USE_JALALI_DATETIME: bool = False

    """
    Using datetime dict or string formatted dicts
    """
    USE_DATETIME_DICT: bool = True

    """
    Date & Datetime string formats
    """
    DATE_STRING_FORMAT = '%Y-%m-%d'
    DATETIME_STRING_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'

    LANGUAGE = LANGUAGES.EN
    NEW_USERS_LANGUAGE = None
    ASYNC_AVAILABLE: bool = False

    # Email Providers
    DJANGO_EMAIL_ENABLE = False
    MAILGUN_EMAIL_ENABLE = False

    # SMS Providers
    KAVENEGAR_SMS_ENABLE = False

    # Django SMTP
    DJANGO_SENDER_ADDRESS: str = None

    # Mailgun
    MAILGUN_DOMAIN_NAME: str = None
    MAILGUN_API_KEY: str = None
    MAILGUN_SENDER_ADDRESS: str = None
    MAILGUN_SENDER_NAME: str = None

    # Kavenegar
    KAVENEGAR_API_TOKEN: str = None
    KAVENEGAR_SIGN_IN_TEMPLATE = None
    KAVENEGAR_SIGN_UP_TEMPLATE = None
    KAVENEGAR_DEFAULT_TEMPLATE = KAVENEGAR_SIGN_IN_TEMPLATE

    # Phone Verification
    PHONE_VERIFICATION_GAP_SECONDS = 90
    PHONE_VERIFICATION_TRIES_COUNT = 3
    PHONE_VERIFICATION_CODE_LENGTH = 4
    PHONE_VERIFICATION_VALID_SECONDS = 10 * 60
    PHONE_VERIFICATION_BODY_STRING = 'Your code is {code}'

    # Email Verification
    EMAIL_VERIFICATION_GAP_SECONDS = 4 * 60
    EMAIL_VERIFICATION_TRIES_COUNT = 3
    EMAIL_VERIFICATION_CODE_LENGTH = 6
    EMAIL_VERIFICATION_VALID_SECONDS = 30 * 60
    EMAIL_VERIFICATION_SUBJECT = 'Email Verification'
    """Only one of these two must be not-null"""
    EMAIL_VERIFICATION_BODY_STRING = 'Your code is {code}'
    EMAIL_VERIFICATION_BODY_HTML = None

    # Email Key Value Authentication
    EMAIL_KEY_VALUE_AUTHENTICATION_VERIFICATION_REQUIRED: bool = True
    EMAIL_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_DOMAIN: str = 'ABCDEFGHKLMNPQRSTUVWXYZabcdefghkmnpqrstuvwxyz23456789'
    EMAIL_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_LENGTH: int = 6
    EMAIL_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_VALID_SECONDS: int = 2 * 60 * 60
    EMAIL_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_GAP_SECONDS: int = 2 * 60 - 10
    EMAIL_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_TRIES_COUNT: int = 8
    EMAIL_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_GAP_SECONDS: int = 2 * 60 - 10
    EMAIL_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_VALID_SECONDS: int = 2 * 60 * 60 - 10
    EMAIL_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_TOKEN_LENGTH: int = 4
    EMAIL_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_TOKEN_DOMAIN: str = '1234567890'
    EMAIL_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_SUBJECT: str = 'Email Verification Reset Password'
    EMAIL_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_BODY: str = 'Reset password token is {token}'
    EMAIL_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_HTML_BODY_TEMPLATE_NAME: str = None

    # Phone Key Value Authentication
    PHONE_KEY_VALUE_AUTHENTICATION_VERIFICATION_REQUIRED: bool = True
    PHONE_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_DOMAIN: str = '1234567890'
    PHONE_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_LENGTH: int = 4
    PHONE_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_VALID_SECONDS: int = 15 * 60
    PHONE_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_GAP_SECONDS: int = 2 * 60 - 10
    PHONE_KEY_VALUE_AUTHENTICATION_VERIFICATION_CODE_TRIES_COUNT: int = 4
    PHONE_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_GAP_SECONDS: int = 2 * 60 - 10
    PHONE_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_VALID_SECONDS: int = 15 * 60 - 10
    PHONE_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_TOKEN_LENGTH: int = 4
    PHONE_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_TOKEN_DOMAIN: str = '1234567890'
    PHONE_KEY_VALUE_AUTHENTICATION_RESET_PASSWORD_SMS_TEMPLATE: str = None

    # Phone OTP
    POA_VERIFICATION_CODE_LENGTH = 4
    POA_VERIFICATION_VALID_SECONDS = 10 * 60
    POA_VERIFICATION_GAP_SECONDS = 60

    # Faker
    FAKER_LOCALE: str = 'fa_IR'
    FAKER_SEED: int = None

    #  Firebase
    FIREBASE_SERVER_TOKEN: Union[str, dict] = ''

    # VisitorToken
    VISITOR_KEY_LENGTH = 40

    # open api
    REQUEST_COMMON_URL_PARAMETERS = [{
        "name": 'language',
        "in": 'query',
        "description": 'set language for this request',
        "required": False,
    }]

    # Neshan
    NESHAN_API_KEY: str = None

    # Chayi
    CHAYI_PROJECT_PACKAGE: str = None
    CHAYI_MODEL_FILE_IMPORTS: str = None

    # Openapi
    OPENAPI_APPLICATION_TITLE = 'NOT_SET'
    OPENAPI_APPLICATION_DESCRIPTION = 'NOT_SET'
    OPENAPI_APPLICATION_VERSION = 'NOT_SET'
    """List of OpenApi.Server"""
    OPENAPI_APPLICATION_SERVERS: list = []

    @classmethod
    def on_startup(cls):
        """
        This method will be called anytime server starts. But just the method from get_avishan_config() result.
        """

        """Any checks needed"""
        # todo
        if cls.PROJECT_NAME is None:
            raise ImproperlyConfigured("PROJECT_NAME must be set in AvishanConfigFather inherited class.")
        if 'avishan' not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured("'avishan' not added to INSTALLED_APPS")
        if 'corsheaders' not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured("'corsheaders' not added to INSTALLED_APPS")
        if 'corsheaders.middleware.CorsMiddleware' != settings.MIDDLEWARE[0]:
            raise ImproperlyConfigured("'corsheaders.middleware.CorsMiddleware' must be in first index of MIDDLEWARE")
        if 'avishan.middlewares.Wrapper' not in settings.MIDDLEWARE:
            raise ImproperlyConfigured("'avishan.middlewares.Wrapper' not added to MIDDLEWARE (usually last item)")
        if 'crum.CurrentRequestUserMiddleware' not in settings.MIDDLEWARE:
            raise ImproperlyConfigured("'crum.CurrentRequestUserMiddleware' not added to MIDDLEWARE "
                                       "(must be before 'avishan.middlewares.Wrapper')")

    @classmethod
    def on_request(cls, request):
        """
        This method called for any request, just before Avishan middleware starts calling get_response()
        """
        pass

    @classmethod
    def get_otp_users(cls) -> List[Type]:
        return []

    @classmethod
    def create_or_update_user_group(cls, title: str, token_valid_seconds: int):
        from avishan.models import UserGroup
        return UserGroup.create_or_update(
            fixed_kwargs={'title': title},
            new_additional_kwargs={
                'token_valid_seconds': token_valid_seconds,
            }
        )

    @classmethod
    def get_country_mobile_numbers_data(cls) -> List[dict]:
        return [
            {
                'name': 'Iran',
                'dialing_code': '98',
                'mobile_number_length': 10,
                'mobile_providers': {
                    'mtn': ['901', '902', '903', '904', '905', '930', '933', '935', '936', '937', '938', '939'],
                    'mci': ['91', '990', '991', '992', '993', '994'],
                    'rightel': ['920', '921', '922'],
                    'mtce': ['931'],
                    'taliya': ['932'],
                    'kish-tci': ['934'],
                    'aptel': ['99910', '99911', '99913'],
                    'azartel': ['99914'],
                    'samantel': ['99999', '99998', '99997', '99996'],
                    'lotustel': ['9990'],
                    'shatel': ['99810', '99811', '99812', '99814'],
                    'ariantel': ['9998'],
                    'anarestan': ['9944']
                }
            }
        ]

    @classmethod
    def get_redoc_schema_models(cls):
        from avishan.models import AvishanModel
        return AvishanModel.get_non_abstract_models()

    @classmethod
    def get_openapi_ignored_path_models(cls) -> List[str]:
        return ['RequestTrackException', 'RequestTrack']

    @classmethod
    def email_key_value_authentication_verification_subject(cls, target=None):
        return 'Email Verification'

    @classmethod
    def email_key_value_authentication_verification_body(cls, target=None):
        return 'Your code is {code}'

    @classmethod
    def email_key_value_authentication_verification_html_body(cls, target=None):
        return None

    @classmethod
    def phone_key_value_authentication_verification_body(cls, target=None):
        return 'Your code is {code}'

    @classmethod
    def email_otp_authentication_verification_subject(cls, target=None):
        return 'Email Verification'

    @classmethod
    def email_otp_authentication_verification_body(cls, target=None):
        return 'Your code is {code}'

    @classmethod
    def email_otp_authentication_verification_html_body(cls, target=None):
        return None

    @classmethod
    def phone_otp_authentication_verification_body(cls, target=None):
        return 'Your code is {code}'

    @classmethod
    def descriptor_ignored_installed_apps(cls) -> List[str]:
        return [
            'django.contrib.admin',
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.messages',
            'django.contrib.staticfiles',
            'corsheaders',
            'dbbackup',
        ]


def get_avishan_config() -> Union[Type[AvishanConfigFather]]:
    for sub_class in AvishanConfigFather.__subclasses__():
        sub_class: Type[AvishanConfigFather]
        return sub_class
    return AvishanConfigFather



from avishan.configure import get_avishan_config
from crum import get_current_request


class AvishanTranslatable:
    EN = None
    FA = None

    # todo 0.2.5: can we translate something by default? like "title", "icon", etc...
    def __init__(self, **kwargs):
        """
        translatable texts
        :param kwargs: keys like: FA, EN
        """

        for key, value in kwargs.items():
            self.__setattr__(key.upper(), value)

    def __str__(self):

        try:
            if get_current_request().avishan.language is None:
                lang = get_avishan_config().LANGUAGE
            else:
                lang = get_current_request().avishan.language
            if lang.upper() in self.__dict__.keys() and self.__dict__[lang.upper()] is not None:
                return self.__dict__[lang.upper()]
        except:
            pass
        try:
            return list(self.__dict__.values())[0]
        except IndexError:
            return 'NOT_TRANSLATED_STRING'

    def __repr__(self):
        return str(self)



# Generated by Django 3.0.7 on 2020-06-14 11:17

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0011_auto_20200526_1740'),
    ]

    operations = [
        migrations.DeleteModel(
            name='RequestTrackExecInfo',
        ),
    ]



from logging import Logger

from celery import shared_task

from avishan.models import PhoneOtpAuthenticate
from celery.utils.log import get_task_logger

logger: Logger = get_task_logger(__name__)


@shared_task(name='avishan.async_phone_otp_authentication_send_otp_code')
def async_phone_otp_authentication_send_otp_code(poa_id: int):
    try:
        PhoneOtpAuthenticate.get(id=poa_id).send_otp_code()
    except PhoneOtpAuthenticate.DoesNotExist:
        logger.error(f'poa {poa_id} not found')



from datetime import timedelta

from django.utils import timezone

from avishan.models import RequestTrack
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Cleans RequestTrack objects'

    def add_arguments(self, parser):
        parser.add_argument(
            '--type',
            type=int,
            help='1: last day & exception-added | 2: last day | 3: exception-added | 4: all',
        )

    def handle(self, *args, **kwargs):
        delete_type = kwargs['type']
        if not delete_type:
            delete_type = 1

        deletes = RequestTrack.objects.all()
        if delete_type in [1, 2]:
            now = timezone.now()
            deletes = deletes.filter(start_time__lte=now - timedelta(days=1))

        if delete_type in [1, 3]:
            deletes = deletes.filter(exception__isnull=True)

        deletes.delete()



# todo 0.2.3: can we count db calls with decorator?
# todo 0.2.2: save user device on every request
# todo 0.2.1: readme edit to latest
# todo 0.2.2: app check function change to class. to use its parameters.
# todo 0.2.2: AvishanWarning with color and text and usage
# todo 0.2.3: execution time avishan_calculate decorator
# todo 0.2.2: video model

"""avishan_config file"""
# todo 0.2.2: create avishan_config.py file to each app: to always run it at startup
# todo 0.2.2: create avishan_config.py 'check' function and 'AvishanConfig' class
# todo 0.2.1: add avishan app to INSTALLED_APPS
# todo 0.2.1: add avishan middlewares to list
# todo 0.2.1: setup datetime settings & use_tz
# todo 0.2.1: setup static & media settings
# todo 0.2.1: handle cors headers
# todo 0.2.1: add and comment database supported settings
# todo 0.2.1: move secret_key to config file
# todo 0.2.1: ask for project name and add it to settings.py. to put avishan_config in project folder not root



# Generated by Django 3.0.9 on 2020-11-23 09:02

import avishan.libraries.faker
import avishan.models_extensions
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0021_auto_20200815_1652'),
    ]

    operations = [
        migrations.CreateModel(
            name='Province',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('country', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='provinces', to='avishan.Country')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, avishan.libraries.faker.AvishanFaker, avishan.models_extensions.AvishanModelDjangoAdminExtension, avishan.models_extensions.AvishanModelModelDetailsExtension, avishan.models_extensions.AvishanModelFilterExtension, avishan.models_extensions.AvishanModelDescriptorExtension),
        ),
        migrations.CreateModel(
            name='City',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('province', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='cities', to='avishan.Province')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, avishan.libraries.faker.AvishanFaker, avishan.models_extensions.AvishanModelDjangoAdminExtension, avishan.models_extensions.AvishanModelModelDetailsExtension, avishan.models_extensions.AvishanModelFilterExtension, avishan.models_extensions.AvishanModelDescriptorExtension),
        ),
    ]
